/*
 * Copyright (C) 2003-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define __MODULES_C__
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rfm_libs.h"
#include "rfm_modules.h"

typedef xfdir_t *(xfdir_f) (record_entry_t * en, widgets_t * widgets_p);

const gchar *
rfm_get_module_popup_id(void){
    return "module_popup_widget";
}

void
rfm_sanity_check (int argc, char **argv, int librfm_serial) {
    if(librfm_serial != LIBRFM_SERIAL) {
        GtkWidget *dialog;
	/* This warning text is only for developers modifying and rebuilding */
        gchar *p = g_strdup_printf ("%s needs to be recompiled \n(has obsolete library headers)", argv[0]);
        gtk_init (&argc, &argv);
        dialog = gtk_message_dialog_new (NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK, "%s", p);
        gtk_dialog_run (GTK_DIALOG (dialog));
        exit (0);
    }
    return;
}

static gint
compare_strings (gconstpointer a, gconstpointer b) {
    return (strcmp ((char *)a, (char *)b));
}

GSList *
rfm_find_root_plugins (void) {
    GDir *dir;
    static GSList *plugin_list = NULL;
    GError *error = NULL;
    gchar *full_libdir = g_build_filename (LIBDIR, PLUGIN_DIR, NULL);

    if((dir = g_dir_open (full_libdir, 0, &error)) != NULL) {
        const gchar *name;
        while((name = g_dir_read_name (dir)) != NULL) {
            gchar *plugin;
            if(strncmp (name, "lib", strlen ("lib")))
                plugin = g_strdup (name);
            else
                plugin = g_strdup (name + strlen ("lib"));
            if(strchr (plugin, '.'))
                plugin = strtok (plugin, ".");

            if(g_slist_find_custom (plugin_list, plugin, compare_strings)) {
                g_free (plugin);
            } else {
                TRACE ("modules: g_slist_append %s\n", plugin);
                plugin_list = g_slist_append (plugin_list, plugin);
            }
        }
        g_dir_close (dir);
    }

    g_free (full_libdir);

    return plugin_list;
}

/*************************************************************/

void unload_module (const gchar * module_name);

GHashTable *module_hash = NULL;
typedef struct module_info_t {
    void *module_functions;
    GModule *module_cm;
} module_info_t;

module_info_t *
get_module_info (const gchar * librarydir, const gchar * module_name) {
    module_info_t *module_info;
    void *(*module_init) (void);
    void *(*module_sanity) (void);
    int init_version;

    TRACE ("getting module info for %s (LIBDIR=%s)\n", module_name, LIBDIR);

    if(!module_hash) {
        TRACE ("get_module_info: creating module hash\n");
        module_hash = g_hash_table_new (g_str_hash, g_str_equal);
        if(!module_hash)
            g_assert_not_reached ();
    }
    TRACE ("looking for module in hash: %s\n", module_name);
    module_info = g_hash_table_lookup (module_hash, module_name);
    if(module_info) {
        TRACE ("module info found (already loaded) 0x%lx\n", (unsigned long)module_info);
        return module_info;
    } else {
        gchar *full_libdir;
        gchar *module;

        if(librarydir)
            full_libdir = g_build_filename (LIBDIR, librarydir, NULL);
        else
            full_libdir = g_strdup (LIBDIR);
        module = g_module_build_path (full_libdir, module_name);
        g_free (full_libdir);
        if(!rfm_g_file_test (module, G_FILE_TEST_EXISTS)) {
            TRACE ("get_module_info: Module %s does not exist!\n", module);
            g_free (module);
            return NULL;
        }

        TRACE ("get_module_info: loading %s\n", module);
        module_info = (module_info_t *) malloc (sizeof (module_info_t));
        if(!module_info)
            g_assert_not_reached ();
        module_info->module_cm = g_module_open (module, 0);
        if(!module_info->module_cm) {
            TRACE ("get_module_info: g_module_open(%s) == NULL\n", module);
            TRACE ("get_module_info: module cannot be opened: %s\n    get_module_info: Check if correctly installed or unresolved symbols within...\n****\n", module);
            g_free (module);
            g_free (module_info);
            return NULL;
        }
        /* module sanity */
        if(!g_module_symbol (module_info->module_cm, "module_sanity", (gpointer) & (module_sanity))) {
            g_warning ("Module %s is not sane!\nDoes module specify: \"LIBRFM_MODULE\"?", module);
            g_free (module);
            g_free (module_info);
            return NULL;
        } else {
            init_version = GPOINTER_TO_INT ((*module_sanity) ());
            if(init_version != LIBRFM_SERIAL) {
                TRACE ("Module %s is compiled with obsolete headers (not loaded)", module);
                g_free (module);
                g_free (module_info);
                return NULL;
            }
        }
        if(!g_module_symbol (module_info->module_cm, "module_init", (gpointer) & (module_init))) {
            TRACE ("g_module_symbol(module_init) != FALSE\n");
            g_free (module);
            g_free (module_info);
            return NULL;
        }

        TRACE ("adding module to module hash: %s (0x%lx)\n", module_name, (unsigned long)module_info);
        module_info->module_functions = (*module_init) ();
        g_hash_table_insert (module_hash, (gpointer) module_name, module_info);
        TRACE ("get_module_info: module %s successfully loaded\n", module);
        g_free (module);
        return module_info;
    }
}


void *
rfm_void (const gchar * librarydir, const gchar * module_name, const gchar * function_id) {
    gchar *(*function) (void);
    module_info_t *module_info;
    if(!librarydir || !module_name || !function_id)
        return NULL;
    module_info = get_module_info (librarydir, module_name);
    if(!module_info)
        return NULL;

    if(!g_module_symbol (module_info->module_cm, function_id, (gpointer) & (function))) {
        /*if (strcmp(function_id,"submodule_name")!=0 && strcmp(function_id,"submodule_dir")!=0) */
        {
            TRACE ("g_module_symbol(%s) != FALSE in module %s\n", function_id, module_name);
        }
        /*unload_module(module_name); */
        return NULL;
    }
    return (*function) ();
}

void *
rfm_natural (const gchar * librarydir, const gchar * module_name, void *n, const gchar * function_id) {
    gchar *(*function) (void *n);
    module_info_t *module_info;
    if(!librarydir || !module_name || !function_id){
	TRACE("!librarydir (%s)|| !module_name (%s)|| !function_id(%s)\n",
		librarydir, module_name, function_id);
        return NULL;
    }
    module_info = get_module_info (librarydir, module_name);
    if(!module_info){
	TRACE("!module_info\n");
        return NULL;
    }
    if(!g_module_symbol (module_info->module_cm, function_id, (gpointer) & (function))) {
        if(strcmp (function_id, "submodule_name") != 0 && strcmp (function_id, "submodule_dir") != 0) {
            TRACE ("g_module_symbol(%s) != FALSE in module %s\n", function_id, module_name);
        }
        /*unload_module(module_name); */
        return NULL;
    }
    return (*function) (n);
}

void *
rfm_rational (const gchar * librarydir, const gchar * module_name, void *p, void *q, const gchar * function_id) {
    gchar *(*function) (void *p, void *q);
    module_info_t *module_info;
    if(!librarydir || !module_name || !function_id)
        return NULL;
    module_info = get_module_info (librarydir, module_name);
    if(!module_info)
        return NULL;
    if(!g_module_symbol (module_info->module_cm, function_id, (gpointer) & (function))) {
        if(strcmp (function_id, "submodule_name") != 0 && strcmp (function_id, "submodule_dir") != 0) {
            TRACE ("g_module_symbol(%s) != FALSE in module %s\n", function_id, module_name);
        }
        /*unload_module(module_name); */
        return NULL;
    }
    return (*function) (p, q);
}

void *
rfm_complex (const gchar * librarydir, const gchar * module_name, void *p, void *q, void *r, const gchar * function_id) {
    gchar *(*function) (void *p, void *q, void *r);
    module_info_t *module_info;
    if(!librarydir || !module_name || !function_id)
        return NULL;
    module_info = get_module_info (librarydir, module_name);
    if(!module_info)
        return NULL;
    if(!g_module_symbol (module_info->module_cm, function_id, (gpointer) & (function))) {
        if(strcmp (function_id, "submodule_name") != 0 && strcmp (function_id, "submodule_dir") != 0) {
            TRACE ("g_module_symbol(%s) != FALSE in module %s\n", function_id, module_name);
        }
        /*unload_module(module_name); */
        return NULL;
    }
    return (*function) (p, q, r);
}

void *
load_module (const gchar * librarydir, const gchar * module_name) {
    module_info_t *module_info;
    module_info = get_module_info (librarydir, module_name);
    TRACE ("finished load attempt of module %s (%s)\n", module_name, librarydir);
    if(!module_info) {
        g_warning ("Show stopper: cannot get module info for %s", module_name);
        exit (0);
    }
    return module_info->module_functions;
}

void
unload_module (const gchar * module_name) {
    module_info_t *module_info;

    if(!module_hash)
        return;
    module_info = g_hash_table_lookup (module_hash, module_name);
    if(!module_info) {
        TRACE ("module %s is not loaded\n", module_name);
        return;
    }

    if(!g_module_close (module_info->module_cm)) {
        g_warning ("g_module_close (%s) failed\n", module_name);
    } else {
        if(!g_hash_table_remove (module_hash, module_name)) {
            g_warning ("could not remove %s from module hash", module_name);
        }
        g_free (module_info);
        TRACE ("module %s unloaded\n", module_name);
    }
}

/*********************************************************************************************/


