/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rfm_libs.h"
#include "rfm_modules.h"
#include "primary-run.i"

// Use rfm_thread_run_argv or rfm_thread_run?
// rfm_thread_run_argv will execute directly, not via a shell, and will
// not be saved in the lpterm history

// rfm_thread_run_argv:
// This modality will execute the command without shell

void
rfm_thread_run2argv (widgets_t * widgets_p, const gchar * command, gboolean interm) {
    // do the call with argv so that command not saved in lpterm history
    gint argcp;
    gchar **argvp;
    if(g_shell_parse_argv (command, &argcp, &argvp, NULL)) {
        TRACE ("OPEN: rfm_thread_run_argv\n");
        // this should always work
        rfm_thread_run_argv (widgets_p, argvp, interm);
        g_strfreev (argvp);
    } else {
        g_warning ("failed to parse command with g_shell_parse_argv() at run.c");
        rfm_thread_run (widgets_p, command, interm);
    }
}

void
rfm_thread_run_argv (
	widgets_t * widgets_p, 
	gchar ** argv, 
	gboolean interm
	){
    private_rfm_thread_run_argv(widgets_p, argv, interm, NULL, NULL);
}

void
rfm_thread_run_argv_with_stdout (
	widgets_t * widgets_p, 
	gchar ** argv, 
	gboolean interm, 
	void (*stdout_f) (void *stdout_data,
                      void *stream,
                      int childFD)
	){
    private_rfm_thread_run_argv(widgets_p, argv, interm, stdout_f, NULL);
}

void
rfm_thread_run_argv_with_stderr (
	widgets_t * widgets_p, 
	gchar ** argv, 
	gboolean interm, 
	void (*stderr_f) (void *stderr_data,
                      void *stream,
                      int childFD)
	){
    private_rfm_thread_run_argv(widgets_p, argv, interm, NULL, stderr_f);
}
// this function is run in a gthread at startup. 
// mutex is to protect from main thread running the
// save history function.
GMutex *command_history_mutex=NULL;
gpointer
rfm_load_sh_command_history (gpointer data) {
    // This thread is not barrier subject
    if (!command_history_mutex) {
	command_history_mutex=g_mutex_new();
    }
    view_t *view_p = (view_t *) data;
    gchar *history = g_build_filename (g_get_user_cache_dir (), LP_TERMINAL_HISTORY, NULL);
    TRACE ("HISTORY: %s\n", history);
    g_mutex_lock(command_history_mutex);
    g_list_free (view_p->sh_command);
    GList *p;
    for(p = view_p->sh_command; p; p = p->next) {
        g_free (p->data);
    }
    view_p->sh_command = NULL;
    view_p->sh_command = g_list_append (view_p->sh_command, g_strdup (""));
    view_p->sh_command_counter = 0;

    FILE *sh_history = fopen (history, "r");
    if(sh_history) {
        // we should put a file lock here...
        TRACE ("HISTORY: readlock %s\n", history);
        rfm_lock_fd_read (fileno (sh_history));
        char line[2048];
        memset (line, 0, 2048);
        while(fgets (line, 2047, sh_history) && !feof (sh_history)) {
            TRACE ("HISTORY: line %s\n", line);
            if(strchr (line, '\n'))
                *strchr (line, '\n') = 0;
            if(strlen (line) == 0)
                continue;
            // skip invalid commands (except cd):
            if(!MIME_is_valid_command (line)) {
                if(strcmp (line, "cd") != 0 && strncmp (line, "cd ", strlen ("cd ")) != 0) {
                    TRACE ("HISTORY: invalid history command in %s: %s\n", history, line);
                    continue;
                }
            }
	    gchar *newline=compact_line(line);
	    GList *element=find_in_history_list(view_p->sh_command, newline);

	    if (element) { 
		// remove old element
		gchar *data=element->data;
		view_p->sh_command = g_list_remove(view_p->sh_command, data);
		g_free(data);
	    }
	    // put at top of the pile
	    view_p->sh_command =
		    g_list_insert_before (view_p->sh_command, g_list_last (view_p->sh_command), newline);
        }
        rfm_unlock_fd (fileno (sh_history));
        TRACE ("HISTORY: readunlock %s\n", history);
        fclose (sh_history);
        view_p->sh_command_counter = g_list_length (view_p->sh_command) - 1;
    }
    g_free (history);
    g_mutex_unlock(command_history_mutex);
    return NULL;
}

void
rfm_save_sh_command_history (view_t * view_p, const gchar * command) {
    if (!view_p){
	return;
    }
    if (!command_history_mutex) {
	command_history_mutex=g_mutex_new();
    }
    GList *p;
    g_mutex_lock(command_history_mutex);
    p = g_list_previous (g_list_last (view_p->sh_command));
    if(!command || !strlen (command)){
	g_mutex_unlock(command_history_mutex);
        return;
    }
    gchar *command_p = g_strdup (command);
    g_strstrip (command_p);
    // if repeat of last command, skip it.
    if(!p || strcmp (command, (char *)p->data)) {
        view_p->sh_command = g_list_insert_before (view_p->sh_command, g_list_last (view_p->sh_command), command_p);
        //view_p->sh_command = g_list_append(view_p->sh_command, command_p);

        /* don't save to file if invalid command */
        if(!MIME_is_valid_command (command_p)) {
            if(strcmp (command_p, "cd") != 0 && strncmp (command_p, "cd ", strlen ("cd ")) != 0) {
                g_warning ("not saving %s", command_p);
                view_p->sh_command_counter = g_list_length (view_p->sh_command) - 1;
		g_mutex_unlock(command_history_mutex);
                return;
            }
        }
        // here we will rewrite the history, removing any item
        // which duplicate the command about to be saved.
        // This effectively moves the command to the bottom
        // of the list.
        // 

        gchar *history = g_build_filename (g_get_user_cache_dir (), LP_TERMINAL_HISTORY, NULL);

        // read it first to synchronize with other rodent instances
        GList *disk_history = NULL;
        FILE *sh_history = fopen (history, "r");

        if(sh_history) {
            // we should put a file lock here...
            rfm_lock_fd_read (fileno (sh_history));
            char line[2048];
            memset (line, 0, 2048);
            while(fgets (line, 2047, sh_history) && !feof (sh_history)) {
                if(strchr (line, '\n')) {
                    *strchr (line, '\n') = 0;
                }
                if(strcmp (line, command_p) != 0) {
                    disk_history = g_list_append (disk_history, g_strdup (line));
                }
            }
            rfm_unlock_fd (fileno (sh_history));
            fclose (sh_history);
        }
        disk_history = g_list_append (disk_history, g_strdup (command_p));

        sh_history = fopen (history, "w");
        if(sh_history) {
            // we should put a file lock here...
            rfm_lock_fd_write (fileno (sh_history));
            GList *p;
            for(p = g_list_first (disk_history); p && p->data; p = p->next) {
                fprintf (sh_history, "%s\n", (char *)p->data);
                g_free (p->data);
            }
            rfm_unlock_fd (fileno (sh_history));
            fclose (sh_history);
        }
        g_list_free (disk_history);
        g_free (history);
    }
    view_p->sh_command_counter = g_list_length (view_p->sh_command) - 1;
    TRACE ("sh_command_counter=%d\n", view_p->sh_command_counter);
    g_mutex_unlock(command_history_mutex);
    return;
}

// rfm_thread_run:
// This modality will execute the command via "sh -c" and allow pipes and 
// redirection. and will be saved in the lpterm history file

void
rfm_thread_run (widgets_t * in_widgets_p, const gchar * command, gboolean interm) {
    pid_t child;
    gchar *exec_command;
    if(interm) {
        const gchar *term = rfm_what_term ();
        const char *exec_option = rfm_term_exec_option(term);
        exec_command = g_strconcat (term, " ", exec_option, " ", command, NULL);
    } else
        exec_command = g_strdup (command);
    gchar *save_command = g_strdup (exec_command);
    g_strstrip (exec_command);
    if(strncmp (exec_command, "sudo", strlen ("sudo")) == 0 && strncmp (exec_command, "sudo -A", strlen ("sudo -A")) != 0) {
        check_sudo ();
        gchar *nc = g_strdup_printf ("sudo -A %s", exec_command + strlen ("sudo"));
        g_free (exec_command);
        exec_command = nc;
    }



    /*g_shell_parse_argv (exec_command, &argc,&argv,&error);
       if (error){
       g_warning("%s",error->message);
       g_error_free(error);
       g_free(exec_command);
       return;
       } */
    gchar *argv[5];
    gint i=0;
    // dash is broken (for our purposes):
    // it will not pass on SIGSEGV to child processes...
    gchar *bash=g_find_program_in_path("bash");
    if (!bash){
	g_warning("GNU bash not found.");
	bash=g_strdup("/bin/sh");
    }
    argv[i++] = bash;
    argv[i++] = "-c";
    argv[i++] = exec_command;
    argv[i++] = NULL;

    int id=Tubo_id ();
    widgets_t *widgets_p = fallback(in_widgets_p);
    child = thread_run (widgets_p, argv, NULL, NULL);
    g_free(bash);
    if(rfm_diagnostics_is_visible (widgets_p)) {
        gchar *g = g_strdup_printf ("%d> (%d):", id, child);
        rfm_diagnostics (widgets_p, "xffm/stock_execute", NULL);
        //rfm_diagnostics(widgets_p, "xffm/stock_yes", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/command_id", g, NULL);
        //rfm_diagnostics (widgets_p, "xffm_tag/command", " ", exec_command, "\n", NULL);
        g_free (g);
        rfm_diagnostics (widgets_p, "xffm_tag/command", " ", exec_command, "\n", NULL);
    }
    if(child > 0) {
        setup_run_button_thread (widgets_p, exec_command, child);
    }
    // instead of disposing of exec_command, we will save command 
    // in the sh_command history 
    rfm_save_sh_command_history (widgets_p->view_p, save_command);
    g_free (exec_command);
    // do not do this: g_free(save_command); 
    // because save_command gets put into the history GList.
}

#if 0
G_MODULE_EXPORT const gchar *
rfm_get_from_open_history (gchar * path2find, int *in_term) {
    gchar *dbh_file = g_build_filename (g_get_user_data_dir (), OPEN_FILE, NULL);
    GString *gs;
    DBHashTable *d;
    gchar *data;
    gchar *basename = g_path_get_basename (path2find);
    static gchar *in_cmd = NULL;

    TRACE ("Opening DBH file %s\n", dbh_file);
    if((d = dbh_open (dbh_file)) == NULL) {
        if((d = dbh_create (dbh_file, 11)) == NULL) {
            unlink (dbh_file);
            if((d = dbh_create (dbh_file, 11)) == NULL) {
                g_free (basename);
                g_free (dbh_file);
                return NULL;
            }
        }
    }
    g_free (dbh_file);
    gs = g_string_new (basename);
    g_free (basename);
    sprintf ((char *)DBH_KEY (d), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    data = (gchar *) DBH_DATA (d);

    if(!dbh_load (d)) {
        dbh_close (d);
        return NULL;
    }

    if(*data == 0)
        *in_term = 0;
    else
        *in_term = 1;
    data++;
    g_free (in_cmd);
    in_cmd = g_strdup (data);

    dbh_close (d);
    return (const gchar *)in_cmd;
}
#endif

void
rfm_recover_flags (gchar * in_cmd, gboolean * interm, gboolean * hold) {
    DBHashTable *runflags;
    GString *gs;
    int *flags;
    gchar *g = g_build_filename (g_get_user_data_dir (), RUN_FLAG_FILE, NULL);
    if((runflags = dbh_open (g)) == NULL) {
        TRACE ("Cannot open %s\n", g);
        *interm = 0;
        *hold = 0;
        return;
    }
    gs = g_string_new (in_cmd);
    sprintf ((char *)DBH_KEY (runflags), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    flags = (int *)runflags->data;
    dbh_load (runflags);
    *interm = flags[0];
    *hold = flags[1];
    dbh_close (runflags);

    TRACE ("flags recovered from dbh file for %s, interm=%d hold=%d\n", in_cmd, *interm, *hold);
}


void
rfm_try_sudo (widgets_t * in_widgets_p, gchar ** arg, gboolean interm) {
    int i;
    int j=0;
    char *exec_argv[MAX_COMMAND_ARGS];

    TRACE ("PROP: try sudo: \n");
    check_sudo ();
    exec_argv[j++] = "sudo";
    exec_argv[j++] = "-A";
    for(i = 0; arg[i] != NULL && j < MAX_COMMAND_ARGS - 2; i++) {
        TRACE (" %s", arg[i]);
        exec_argv[j++] = arg[i];
    }

    widgets_t *widgets_p = fallback(in_widgets_p);

    rfm_show_text(widgets_p);

    if (j==MAX_COMMAND_ARGS - 1) {
	rfm_show_text(widgets_p);
    	rfm_diagnostics(widgets_p,"xffm/stock_dialog-warning",NULL);
        gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", strerror(E2BIG)," (> ",max,")","\n", NULL);
        g_free(max);
    }

    exec_argv[j++] = NULL;
    TRACE (" \n");

    rfm_thread_run_argv (widgets_p, exec_argv, interm);
    return;
}
