/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rfm_libs.h"
#include "rfm_modules.h"

#include "primary-xfdir.i"

/**
 * SECTION:xfdir
 * @short_description: xfdir structure is a location to place the contents of 
 * the rodent view. The primary xfdir functions allow for the loading and
 * monitoring of local files on a POSIX compliant system.
 * 
 * @section_id:	xfdir
 * @stability: Current
 * @include: primary.h
 *
 **/

/**
 * xfdir_get_local:
 * @view_p: A Rodent view_t pointer
 * @en: An icon entry structure.
 *
 * Generalized xfdir load routine. en->path  should be a valid
 * path or else en->module defined so that module  particulars 
 * are used.
 *
 * This is main thread executed.
 **/

void
rfm_xfdir_make_layout(view_t * view_p, population_t *population_p){
    if (!view_p || !population_p || !population_p->en) return;
    record_entry_t *en=population_p->en;
    if (!en->path || !en->st || !rfm_g_file_test(en->path, G_FILE_TEST_EXISTS)){
	return;
    }
    // Use layout2 for file details (if applicable)
    // This function is used by monitor for updates on stat changes.
    gchar *mode_string=rfm_mode_string (en->st->st_mode);
    gchar *grupo=rfm_group_string(en->st);
    gchar *owner=rfm_user_string(en->st);
    gchar *date_string=rfm_date_string (en->st->st_mtime);

    gchar *layout2_text=g_strdup_printf("   <span><i>%s\t%s:%s %ld\t%s</i></span>", 
	    mode_string,
	    owner, grupo,
	    (long)en->st->st_size,
	    date_string);
    g_free(mode_string);
    g_free(grupo);
    g_free(owner);
    g_free(date_string);
    population_p->layout2 = 
	gtk_widget_create_pango_layout (view_p->widgets.paper, NULL);
    pango_layout_set_markup(population_p->layout2, layout2_text, -1);
}


xfdir_t *
rfm_xfdir_get (view_t * view_p, record_entry_t * en, gboolean fullbones) {
    
    // set the dtfile quick dir buttons (or menu items in our case)
    if(!view_p->module) {
        if(!en || !en->path || !rfm_g_file_test (en->path, G_FILE_TEST_IS_DIR)) {
            g_warning ("Invalid entry sent to xfdir_get_local");
            return NULL;
        }
        set_deepest_dir (view_p);
    }

    // allocate memory for new xfdir structure and initialize values
    xfdir_t *xfdir_p = (xfdir_t *) malloc (sizeof (xfdir_t));
    memset ((void *)xfdir_p, 0, sizeof (xfdir_t));
    strncpy (xfdir_p->path, en->path, 1023);
    xfdir_p->path[1023] = 0;
    xfdir_p->type = en->type;
    xfdir_p->view_p = view_p;
    if (en && en->st) TRACE ("rfm_xfdir_get: en->st->st_uid  %d\n", en->st->st_uid );

    // This limitation is to make things faster since the desktop will refuse 
    // to map whatever does not fit on the desktop.
    if (view_p->flags.type ==DESKVIEW_TYPE){
	xfdir_p->limited_read = TRUE;
    }


    // set user defined sort preferences and do the sort
    rfm_get_view_preferences (view_p, en);
    
    xfdir_p->ascending = view_p->flags.preferences & SORT_ASCENDING;
    xfdir_p->sort_column = view_p->flags.sortcolumn;
    if (xfdir_p->sort_column != TYPE_SORT && xfdir_p->sort_column != NAME_SORT){
	// we will need to stat records here, so the easiest
	// and quickest way is to enable fullbones (which does not
	// add too much response time, even for large directory listings)
	fullbones=TRUE;
    }
    
    // read local directory information and update xfdir structure
    TRACE ("starting time on update_xfdir for %s...\n", xfdir_p->path);
    update_xfdir (xfdir_p, fullbones);


    // now we do the sorting. Stat record is not necessary for
    // name or type sort methods, so barebones should be fine.
    TRACE("---------- sortcolumn=%d\n", xfdir_p->sort_column);
    xfdir_sort (xfdir_p);

    // set default status tag for directory or module (sizetag)
    rfm_set_entry_tag (&(view_p->widgets), en, xfdir_p->tama);

    // create monitor variables mutex if not created already
    if(view_p->mutexes.monitor_id == NULL) {
        view_p->mutexes.monitor_id = g_mutex_new ();
    }
    // start background directory monitor thread
    if (!view_p->module || 
	    (view_p->module && rfm_void(PLUGIN_DIR, view_p->module, "module_monitor"))){
	xfdir_start_monitor (view_p, xfdir_p);
    }

    return xfdir_p;
}

/**
 * rfm_xfdir_exit_monitor:
 * @xfdir_p:  pointer to an xfdir_t structure
 *
 * Schedule monitor to exit as soon as possible. This function should 
 * be accessed from threads other than the monitor thread. For the monitor thread,
 * use the static exit_monitor() function instead which does not have a
 * wait condition.
 **/
void
rfm_xfdir_exit_monitor (view_t * view_p) {
    if(view_p->mutexes.monitor_id == NULL) {
        view_p->mutexes.monitor_id = g_mutex_new ();
    }
    g_mutex_lock (view_p->mutexes.monitor_id);
    view_p->flags.monitor_id++;
    g_mutex_unlock (view_p->mutexes.monitor_id);
}

/**
 * rfm_xfdir_count_files:
 * @path: directory path
 * Returns: file count within directory, including hidden files
 *
 * This function is non-recursive.
 **/
gint
rfm_xfdir_count_files (const gchar * path) {
    if(!rfm_g_file_test (path, G_FILE_TEST_EXISTS))
        return 0;
    if(!rfm_g_file_test (path, G_FILE_TEST_IS_DIR))
        return 0;
    DIR *directory;
    directory = opendir (path);
    if(!directory)
        return 0;
    gint count = 0;
    struct dirent *d;
    while((d = readdir (directory)) != NULL) {
        if(strcmp (d->d_name, ".") == 0)
            continue;
        if(strcmp (d->d_name, "..") == 0)
            continue;
        count++;
    }
    closedir (directory);
    return count;
}

/**
 * rfm_xfdir_count_hidden_files:
 * @path: directory path
 * Returns: hidden file count within directory
 *
 * This function is non-recursive.
 **/
gint
rfm_xfdir_count_hidden_files (const gchar * path) {
    if(!rfm_g_file_test (path, G_FILE_TEST_EXISTS))
        return 0;
    if(!rfm_g_file_test (path, G_FILE_TEST_IS_DIR))
        return 0;
    DIR *directory;
    directory = opendir (path);
    if(!directory)
        return 0;
    gint count = 0;
    struct dirent *d;
    while((d = readdir (directory)) != NULL) {
        if(strcmp (d->d_name, ".") == 0)
            continue;
        if(strcmp (d->d_name, "..") == 0)
            continue;
        if(d->d_name[0] != '.')
            continue;
        count++;
    }
    closedir (directory);
    return count;
}


