/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"
#include "rodent_population.i"

gint
rodent_get_population_serial (view_t * view_p) {
    if(view_p->mutexes.population_serial == NULL) {
        view_p->mutexes.population_serial = g_mutex_new ();
    }
    g_mutex_lock (view_p->mutexes.population_serial);
    gint population_serial = view_p->population_serial;
    g_mutex_unlock (view_p->mutexes.population_serial);
    return population_serial;
}

void
rodent_layout (view_t * view_p, population_t * population_p, const gchar * label) {
    TRACE (">> rodent_layout\n");
    if(!view_p || !population_p) {
        g_warning ("!view_p || !population_p");
        return;
    }

    gchar *actual_tag = rfm_utf_string (label);
    mk_layouts (view_p, actual_tag, &(population_p->layout), &(population_p->layout2));
    g_free (actual_tag);

    mk_logical_rect (population_p->layout, &(population_p->logical_rect));
    mk_logical_rect (population_p->layout2, &(population_p->logical_rect2));
    return;
}

void
rodent_update_cut_icons (view_t * view_p) {
    TRACE ("CUT:update_cut_status\n");
    if(!rfm_population_try_read_lock (view_p))
        return;
    // find and update record
    population_t **tmp;
    for(tmp = view_p->population_pp; *tmp; tmp++) {
        population_t *population_p = *tmp;
        gboolean expose = FALSE;
        if(population_p->en && population_p->en->path) {
            int pasteboard_status = rfm_in_pasteboard (view_p, population_p->en);
            if(pasteboard_status) {
		if (pasteboard_status==2) { //cut
		    if (!(population_p->flags & POPULATION_CUT)) {
			expose = TRUE;
		    }
		    population_p->flags |= POPULATION_CUT;
		    population_p->flags &= (POPULATION_COPIED ^ 0xffffffff);

		} else {
		    if (!(population_p->flags & POPULATION_COPIED)) {
			expose = TRUE;
		    }
		    population_p->flags |= POPULATION_COPIED;
		    population_p->flags &= (POPULATION_CUT ^ 0xffffffff);
		}
                TRACE ("CUT: pasteboard: %s\n", population_p->en->path);
            } else {
                TRACE ("CUT: not in pasteboard: %s\n", population_p->en->path);
		if (population_p->flags & (POPULATION_COPIED|POPULATION_CUT)) {
		    expose = TRUE;
		}
		population_p->flags &= (POPULATION_COPIED ^ 0xffffffff);
		population_p->flags &= (POPULATION_CUT ^ 0xffffffff);
            }
            if(expose){
		rodent_expose_item (view_p, population_p);
	    }
        }
    }

    // release population semaphore
    TRACE ("POP_SEM:  sem_post(&(view_p->population_sem))\n");
    rfm_population_read_unlock (view_p);
}

static void
set_population_grid_row (view_t * view_p, population_t *population_p, gint element){
    if (view_p->flags.type == ICONVIEW_TYPE) {
	if(view_p->view_layout.grid_columns) {
	    population_p->row = element / view_p->view_layout.grid_columns;
	    population_p->column = element % view_p->view_layout.grid_columns;
	} else {
	    population_p->row = 0;
	    population_p->column = element;
	}
    } else { // transposed matrix for desktop
	if(view_p->view_layout.grid_rows) {
	    population_p->row = element % view_p->view_layout.grid_rows;
	    population_p->column = element / view_p->view_layout.grid_rows;
	} else {
	    population_p->row = element;
	    population_p->column = 0;
	}
    }
}

void
rodent_recalc_population_grid(view_t * view_p){
    
    if (!view_p || !view_p->population_pp || !view_p->population_pp[0]) return;
    if(!rfm_population_try_read_lock (view_p))
        return;
   rodent_init_grid (view_p, view_p->view_layout.max_elements);

    population_t **population_pp=view_p->population_pp;
    gint element=0; 
    for (; population_pp && population_pp[element]; element++){
	population_t *population_p = population_pp[element];
	set_population_grid_row(view_p, population_p, element);
    }
    rfm_population_read_unlock (view_p);
}

/**
 *
 * rodent_create_population_t: create a population element
 *
 * @view_p:
 * @en: the entry to assign to the population element.
 * @element: the element serial id number 
 * @icon_id: the mime_icon identifier. If NULL, then icon is obtained from entry information.
 * @label: the text which is to appear under the icon. If NULL, then the text is inferred from entry information.
 * 
 * **/

population_t *
rodent_create_population_t (view_t * view_p,
	record_entry_t * en, 
	gint element, 
	const gchar * icon_id, 
	const gchar * label) 
{
    GdkPixbuf *pixbuf = NULL;
    population_t *population_p;

    population_p = (population_t *) malloc (sizeof (population_t));
    memset (population_p, 0, sizeof (population_t));
    population_p->view_p = view_p;
	
    set_population_grid_row(view_p, population_p, element);

    population_p->serial = view_p->flags.serial++;
    ///////////////////////////////////////

    if(en) {
        population_p->en = rfm_copy_entry (en);
	//  population_p->en->module = view_p->submodule_name;
    } else {
        population_p->en = NULL;
    }
    TRACE("rodent_create_population_t: path=%s: module=%s\n", 
	    (population_p->en)?population_p->en->path:"NULL",
	    (population_p->en)?population_p->en->module:"NULL");


    population_p->flags = 0;
    population_p->pixbuf = get_population_pixbuf(view_p, population_p, icon_id, element);
    TRACE("get_population_pixbuf done\n"); 
    int pasteboard_status = rfm_in_pasteboard (view_p, population_p->en);
    TRACE("rfm_in_pasteboard done\n"); 
    if(pasteboard_status == 2) {
	population_p->flags |= POPULATION_CUT;
    } else if (pasteboard_status == 1) {
	population_p->flags |= POPULATION_COPIED;
    }

    if(GDK_IS_PIXBUF(pixbuf)) {
        population_p->pixbufW = gdk_pixbuf_get_width ((const GdkPixbuf *)pixbuf);
        population_p->pixbufH = gdk_pixbuf_get_height ((const GdkPixbuf *)pixbuf);
    } else {
        population_p->pixbufW = ICON_SIZE(view_p);
        population_p->pixbufH = ICON_SIZE(view_p);
    }

    /*population_p->submodule_name=NULL; */


    rodent_layout (view_p, population_p, label);
    TRACE("rodent_layout done\n"); 

    // Layout2 is not used when iconsize >= TINY_ICON_SIZE && iconsize < SMALL_ICON_SIZE
    // Layout2 is used for ls -l info when iconsize < TINY_ICON_SIZE
    if (view_p->view_layout.icon_size < TINY_ICON_SIZE){
	if (population_p->layout2){
	    g_error("rodent_population.c: population_p->layout2 memory leak");
	}
	if (view_p->en && view_p->en->module && 
		rfm_rational(PLUGIN_DIR,view_p->en->module, NULL, NULL, "make_list_layout2")){
		rfm_rational(PLUGIN_DIR,view_p->en->module, view_p, population_p, "make_list_layout2");
	}
	else {
	    rfm_xfdir_make_layout(view_p, population_p);
	}
    }

    population_p->rect.x = population_p->column * CELLWIDTH(view_p);
    population_p->rect.y = population_p->row * CELLHEIGHT(view_p);
    population_p->rect.width = CELLWIDTH(view_p);
    population_p->rect.height = CELLHEIGHT(view_p);

    return population_p;
}

void
rodent_destroy_population (view_t * view_p) {
    if(!view_p || !view_p->population_pp)
        return;

    population_t **tmp;
    TRACE ("rodent_population: rodent_destroy_population\n");
    increase_population_serial (view_p);
    for(tmp = view_p->population_pp; *tmp; tmp++) {
        population_t *population_p = *tmp;
        destroy_population_item (population_p);
        population_p = NULL;
    }

    /*memset((void *)(view_p->population_pp),
       0,
       view_p->view_layout.max_elements*sizeof(population_t *)); */

    g_free (view_p->population_pp);
    view_p->population_pp = NULL;
}

population_t **
rodent_mk_grid_elements (view_t * view_p, int grid_elements) {
    TRACE (">> graphics_mk_grid_elements\n");
    population_t **population_pp;
    /* we use (grid_elements+1) to use a NULL pointer as an
     * end of list marker */
    TRACE ("grid_elements=%d (elements)\n", grid_elements);
    increase_population_serial (view_p);        // at destroy and here for transfer too.
    population_pp = (population_t **) malloc ((grid_elements + 1) * sizeof (population_t *));
    memset ((void *)(population_pp), 0, (grid_elements + 1) * sizeof (population_t *));
    population_pp[grid_elements] = NULL;
    return population_pp;
}

record_entry_t *
rodent_init_view (view_t * view_p, record_entry_t * en) {
    TRACE ("rodent_population: rodent_init_view\n");

    // destroy obsolete entry:
    if(view_p->en) rfm_destroy_entry (view_p->en);
    rodent_destroy_population (view_p);
    view_p->mouse_event.boxX = -1, view_p->mouse_event.boxY = -1;
    view_p->mouse_event.old_X = -1, view_p->mouse_event.old_Y = -1;
    if(view_p->selection_list) {
	GSList *tmp=view_p->selection_list;
	for (;tmp && tmp->data; tmp=tmp->next){
	    record_entry_t *en=tmp->data;
	    rfm_destroy_entry(en);
	}
        g_slist_free (view_p->selection_list);
        view_p->selection_list = NULL;
    }
    view_p->mouse_event.mouseX = -1;
    view_p->mouse_event.mouseY = -1;
    view_p->tip_event.tip_timer = 0;
    view_p->widgets.rename = NULL;
    view_p->mouse_event.label_p = NULL;
    view_p->mouse_event.doing_drag_p = NULL;
    view_p->mouse_event.saturated_p = NULL;
    view_p->mouse_event.selected_p = NULL;
    view_p->en = en;

    view_p->xfdir_p = NULL;

    return en;
}


#define TIMERVAL 150

static gboolean updating_pixbuf_hash = FALSE;
static void *
update_pixbuf_hash_f(void *data){
    if (updating_pixbuf_hash) return NULL;
    updating_pixbuf_hash=TRUE;
    GtkWidget *window=data;
    DBG("---->   time to update pixbuf hash!\n");
    //gdk_flush();
    //Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
    //XSync(display, False);  
    rfm_change_icontheme(window);
    //sleep(1);
    updating_pixbuf_hash=FALSE;
    return NULL;
}

gint
rodent_monitor (gpointer data) {

    static gboolean rodent_monitor_active=FALSE;
    // make no mistake, only one monitor and in the main thread
    if (rodent_monitor_active) return TRUE;
    rodent_monitor_active=TRUE;
    GSList *tmp;
    static GThread *valid_thread = NULL;
    static gint timer = 0;
    if(!timer) {
        timer = g_timeout_add (TIMERVAL, rodent_monitor, data);
        valid_thread = (GThread *) data;
	rodent_monitor_active=FALSE;
        return TRUE;
    }

    /* thread protection (just in case, though it does not seem necessary) */
    GThread *self = g_thread_self ();
    if(self != valid_thread) {
        g_warning
            ("Dropping invalid access attempt to timeout function 0x%lu != 0x%lu\n",
             (long unsigned)self, (long unsigned)valid_thread);
	rodent_monitor_active=FALSE;
        return TRUE;
    }

    TRACE ("rodent_monitor, calling thread 0x%lu\n", (unsigned long)self);

    rfm_global_p->window_view_list_readers++;
    static gchar *use_gtk_theme_icons = NULL;
    if (!use_gtk_theme_icons){
	if (getenv("RFM_USE_GTK_ICON_THEME")){
	    use_gtk_theme_icons = 
	       g_strdup(getenv("RFM_USE_GTK_ICON_THEME"));
	} else {
	    use_gtk_theme_icons=g_strdup("");
	}
    }
    const gchar *instant_gtk_theme_icons =
	(getenv("RFM_USE_GTK_ICON_THEME"))?
	   getenv("RFM_USE_GTK_ICON_THEME"):"";

    if (strcmp(use_gtk_theme_icons,instant_gtk_theme_icons) != 0){
	view_t *view_p = rfm_global_p->window_view_list->data;
	GtkWidget *window = view_p->widgets.window;
	g_free(use_gtk_theme_icons);
	use_gtk_theme_icons = g_strdup(instant_gtk_theme_icons);
	//Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
	//XSync(display, False);   
	g_thread_create(update_pixbuf_hash_f, window, FALSE, NULL);
    }
    for(tmp = rfm_global_p->window_view_list; tmp; tmp = tmp->next) {
        view_t *view_p = (view_t *) tmp->data;
        if(view_p->flags.status == STATUS_EXIT){
            continue;
	}
	// if the view is busy with a reload, do not do anything
	if(view_p->flags.status != STATUS_IDLE)
            continue;
	if (view_p->flags.protected){
            continue;
	}
        if(view_p->mutexes.monitor_loop == NULL) {
            view_p->mutexes.monitor_loop = g_mutex_new ();
        }
        if(!view_p->xfdir_p)
            continue;

        if(!view_p || !view_p->en) {
            TRACE ("gridview_monitor:  !view_p || !view_p->en... no thread monitor.\n");
            continue;
        }
	g_mutex_lock(view_p->mutexes.janitor);
	view_p->janitor.monitors++;
	g_mutex_unlock(view_p->mutexes.janitor);

        if(view_p->flags.refresh) {
            TRACE ("gridview_monitor:  view_p->xfdir_p->refresh\n");
            TRACE ("monitor refresh xfdir_p 0x%lx != xfdir_p->xfdir_p 0x%lx\n", (unsigned long)((long)view_p->xfdir_p),
                 (unsigned long)((long)view_p->xfdir_p->xfdir_p));
	    // ignore refresh conditions if any read_lock is set
	    if (view_p->population_lock_count == 0) {
		 if (rodent_reload_view ((gpointer) view_p)){
		    view_p->flags.refresh = 0;
		 }
	    } else {
		TRACE("Cannot refresh, view_p->population_lock_count=%d\n",
			view_p->population_lock_count);
	    }
        }
 	g_mutex_lock(view_p->mutexes.janitor);
	view_p->janitor.monitors--;
	g_mutex_unlock(view_p->mutexes.janitor);
   }
    rfm_global_p->window_view_list_readers--;
    rodent_monitor_active=FALSE;
    return TRUE;
}
// rodent_destroy_view: destroy a single view.
gboolean
rodent_destroy_view_item (view_t * view_p) {
    TRACE (">> rodent_destroy_view_item\n");
    widgets_t *widgets_p = &(view_p->widgets);
    // neutralize diagnostics:
    if (widgets_p->diagnostics_window==NULL && 
	    widgets_p->diagnostics) {
	*(widgets_p->diagnostics)=NULL;
    }

   // First, hide all widgets.
    rfm_global_p->window_view_list_readers++;
    if (g_slist_length(rfm_global_p->window_view_list) == 1) {
	TRACE("hiding single view\n");
	    gtk_widget_hide(widgets_p->window);
    }

    else {
      if (widgets_p->notebook && *(widgets_p->notebook)) {
    	//gtk_widget_hide(view_p->widgets.page_child);
	gint page_num = gtk_notebook_page_num (GTK_NOTEBOOK (*(widgets_p->notebook)), widgets_p->page_child);
	gtk_notebook_remove_page (GTK_NOTEBOOK (*(widgets_p->notebook)), page_num);    
       } else {
    	gtk_widget_hide(widgets_p->window);
       }
    }
    rfm_global_p->window_view_list_readers--;

    // Tag view for cleanup by janitor.
    rfm_xfdir_exit_monitor (view_p);
    view_p->flags.status = STATUS_EXIT;

    return TRUE;
}

// rodent_destroy_view: destroys all views in the current window.
gboolean
rodent_destroy_view (view_t * view_p) {
    TRACE (">> rodent_destroy_view\n");
   // First, hide all widgets.
    gtk_widget_hide(view_p->widgets.window);

    rfm_global_p->window_view_list_readers++;
    GSList *list=rfm_global_p->window_view_list;
    for (;list && list->data; list=list->next){
    	TRACE("destroying view 0x%x \n",
		GPOINTER_TO_INT(list->data));
	rodent_destroy_view_item((view_t *)list->data);
    }
    rfm_global_p->window_view_list_readers--;
    return FALSE;
}

void
rodent_add_view (view_t * view_p) {
    rfm_global_p->window_view_list_readers++;
    if (!g_slist_find (rfm_global_p->window_view_list, view_p)){
	gboolean start_janitor = (rfm_global_p->window_view_list==NULL);
	rfm_global_p->window_view_list = 
	    g_slist_append (rfm_global_p->window_view_list, view_p);
	if (start_janitor){
	    TRACE("Starting janitor thread...\n");
	    g_thread_create(janitor, view_p->widgets.window, FALSE, NULL);
	}
    }
    rfm_global_p->window_view_list_readers--;
}

 // generalized short reload routine...
 // when this is triggered, monitor is in or about to enter the wait state
 // this is a soft reload.
gboolean
rodent_reload_view (gpointer data) {
    view_t *view_p = (view_t *) data;
    if(view_p->flags.status == STATUS_EXIT) return FALSE;
    if (view_p->flags.protected){
	// XXX: for Delta
	// This flag should not be necessary when reload is on fully
	// threaded version and no gtk main iterations are present.
	// This flag protects other code spots too, but it is too
	// difficult to identify all necessary spots. Hence the
	// need to threaditize the function. Threaditization is
	// necessary to have the progress bar working and related
	// feedback to user on long loads.
	TRACE("invalid recursive call attempted on view_p->flags.protected\n");
	return 0;
    } else {
	view_p->flags.protected=TRUE;
    }
	
    rfm_barrier_stop_or_increment();

    //if (!view_p->xfdir_p || !view_p->xfdir_p->xfdir_p) {
    if(!view_p->xfdir_p) {
        g_error ("short_rodent_reload_view: !view_p->xfdir_p");
    }
    TRACE ("starting short_rodent_reload_view...\n");
    /*  monitor in wait state  with view_p->flags.refresh */

    TRACE ("POP_SEM: reload_view lock\n");
    gint serial=view_p->population_serial;
    rfm_population_write_lock (view_p);
    TRACE ("population_sem: reload_view() obtained...\n");
    if (serial != view_p->population_serial){
	TRACE("soft reload is aborting. Population serial is %d (soft reload at %d)\n",
		view_p->population_serial, serial);
	rfm_population_write_unlock (view_p);
	rfm_barrier_decrement();
	return FALSE;
    }
    // population semaphore is ours and population serial is valid...

    /* make new population_pp and transfer old population_pp elements,
     * creating new elements for those not in old population_pp
     * and eliminating those in old population_pp no longer present
     * */
    transfer_population (view_p);
    rodent_init_grid (view_p, view_p->xfdir_p->pathc);
    
    if(view_p->flags.type == ICONVIEW_TYPE) {
	TRACE ("init_grid at short_reload...\n");
        // this uses initgrid info but also may generate an expose
        // so must be placed after transfer_population()
        rodent_restore_view_size (view_p);
    } 
    // recalculate x,y
    // get x,y coordinates for each population_p
    // or correct row, column
    rodent_recalc_population_geometry (view_p);
 
    if(view_p->en && view_p->en->st)
        memcpy (view_p->en->st, &view_p->xfdir_p->st, sizeof (struct stat));
    gdk_flush ();

    // fire off pending preview threads, (thread managed):
    TRACE("softreload: rfm_fireup_previews\n");
    rfm_fireup_previews (view_p);
    

    TRACE ("short_reload_view: reload is done\n");
    view_p->flags.status = STATUS_IDLE;
    // this is not necesary here, reload condition
    // generated by monitor has left monitor in wait stat
    // view_p->xfdir_p->active = 0;
    TRACE ("POP_SEM:  main thread sem_post\n");
    const population_t *population_p = rodent_find_in_population (view_p, view_p->mouse_event.current_mouseX, view_p->mouse_event.current_mouseY);
    if (population_p) {
	rodent_saturate_item (view_p, population_p);	
    }

    rodent_expose_all (view_p);
    rfm_population_write_unlock (view_p);
    TRACE ("population_sem: reload_view(): released!\n");
    view_p->flags.protected=FALSE;
    rfm_barrier_decrement();
    return TRUE;

}

//*** end progress bar stuff ***/

void
rodent_add_icon_roots (view_t * view_p) {
    population_t *population_p;
    GSList *list,
     *plugin_list;

    TRACE ("call to rodent_add_icon_roots\n");
    plugin_list = rfm_find_root_plugins ();
    gint max_elements = 0;
    for(list = plugin_list; list; list = list->next) {
	if (test_root_element (view_p, list->data)) max_elements++;
    }
    max_elements+=2;

    TRACE ("ROOT: rodent_add_icon_roots: gdk_window_resize(%d,%d) rows=%d, cols=%d\n",
         view_p->view_layout.paperX, view_p->view_layout.paperY, 
	 view_p->view_layout.grid_rows, view_p->view_layout.grid_columns);
    if (view_p->flags.type != DESKVIEW_TYPE) { 
      gtk_widget_set_size_request (view_p->widgets.paper, view_p->view_layout.paperX, view_p->view_layout.paperY);
    }      
    view_p->population_pp = rodent_mk_grid_elements (view_p, max_elements);

    TRACE("max_elements = %d\n",max_elements);


     /*init_icon_view readjusts rows if necessary */
    rodent_init_grid (view_p, max_elements);
     
    TRACE(">>>  column_width is now %d\n", view_p->view_layout.column_width);


    gint item=0;
    /* create filesystem root icon */
    population_p = view_p->population_pp[item] = create_local_root (view_p, item, NULL);
    item++;
    /* create homedir icon */
    population_p = view_p->population_pp[item] = create_local_root (view_p, item, g_get_home_dir ());
    TRACE ("homedir icon column, row = %d, %d\n",population_p->column, population_p->row);
    item++;

    TRACE ("ROOTS: now plugins...\n");
    for(list = plugin_list; list; list = list->next) {
        TRACE ("ROOTS: creating population item for module %s\n", (char *)list->data);
        population_p = view_p->population_pp[item] = create_icon_root_element (view_p, item, list->data);
	if (!population_p) continue;
	TRACE ("%s column, row = %d, %d\n",(char *)list->data,
		population_p->column, population_p->row);
        item++;
    }
        rodent_recalc_population_geometry (view_p);

    return;
}

//
// Full reload: this is a hard reload.
//     end icon roots stuff
int
rodent_full_reload_view (gpointer data, record_entry_t * target_en) {
    view_t *view_p = (view_t *) data;
    widgets_t *widgets_p = &(view_p->widgets);
    
    TRACE ("rodent_full_reload_view... %s\n", 
	    (target_en && target_en->path)?target_en->path:"NULL");
    if (view_p->flags.protected){
	TRACE("invalid recursive call attempted on view_p->flags.protected\n");
	return 0;
    } else {
	TRACE("protecting\n");
	view_p->flags.protected=TRUE;
    }

    rfm_barrier_stop_or_increment();
    
retry_reload:
    TRACE ("figuring out best_iconsize(view_p)\n");
    TRACE("rodent_population.c: rfm_get_view_preferences \n");
     
    if (view_p->flags.thumbnailer_active) {
	view_p->flags.thumbnailer_exit=TRUE;    
    }
   
    DBG("1.view_p->view_layout.icon_size = %d\n", view_p->view_layout.icon_size);
    // get preferences for the particular entry
    rfm_get_view_preferences (view_p, target_en);
    DBG("2.view_p->view_layout.icon_size = %d\n", view_p->view_layout.icon_size);

    // save desktop directory for navigational desktop
    save_environment_desktop_dir(view_p, target_en);

    TRACE ("starting rodent_full_reload_view... protected=%d\n", view_p->flags.protected);
    
    // set the icon size scale for gridview windows
    set_icon_size_scale(view_p);
    DBG("size for view determined...%d\n", view_p->view_layout.icon_size);

    // if we claim write lock here, other gtk events which
    // require read lock will have to wait. That's cool.
    // This write lock request blocks until write lock granted
    // (all threads with readlocks must not deadlock before releasing lock!)
    // First we tell other threads that may be doing useless work to quit and
    // release the lock:
    view_p->flags.user_wants_lock=TRUE;
    rfm_population_write_lock (view_p);
    view_p->flags.user_wants_lock=FALSE; //she's got it.

    // past this point, the reload is nonstop.
    // schedule termination of any running monitor threads:
    rfm_xfdir_exit_monitor (view_p);
    rfm_cursor_wait (view_p->widgets.window);

    // init_view will destroy population, so we must have 
    // obtained the write lock by here.   
    rodent_init_view (view_p, target_en);
    if(view_p->flags.type != DESKVIEW_TYPE) {
        rodent_set_view_title (view_p);
    }

    // Rodent root scenario (entry is NULL)
    if(!view_p->en) {
        TRACE("/* root icons special case scenario */\n");
	if(view_p->selection_list) {
	    GSList *tmp=view_p->selection_list;
	    for (;tmp && tmp->data; tmp=tmp->next){
		record_entry_t *en=tmp->data;
		rfm_destroy_entry(en);
	    }
	    g_slist_free (view_p->selection_list);
	    view_p->selection_list = NULL;
	}
        rfm_cursor_reset (view_p->widgets.window);
        view_p->flags.status = STATUS_IDLE;
        if(GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window)) {
	    gtk_adjustment_set_value (
		gtk_scrolled_window_get_vadjustment (
		    view_p->widgets.scrolled_window), 0);
	}
	rodent_add_icon_roots (view_p);
        rfm_population_write_unlock (view_p);
	view_p->module=FALSE;
	view_p->flags.protected=FALSE;
	const population_t *population_p = 
	    rodent_find_in_population (view_p, view_p->mouse_event.current_mouseX, view_p->mouse_event.current_mouseY);
	if (population_p) {
	    rodent_saturate_item (view_p, population_p);	
	}
	rfm_status (widgets_p, "xffm/device_computer", _("Installed Plugins"), NULL);

        rodent_expose_all (view_p);
	rfm_barrier_decrement();

       return TRUE;
    }

    view_p->module = target_en->module;

    // Apply shows hidden flag for specific entry
    if (SHOWS_HIDDEN(view_p->flags.preferences)) {
	SET_SHOWS_HIDDEN(target_en->type);
    } else {
	UNSET_SHOWS_HIDDEN(target_en->type);
    }


    // Thread 1: load the xfdir structure.
    ///////////////////////////////////////////////////
    /* We now load the xfdir structure (there is none for roots). */
    // doing load_xfdir_structure() by thread blocks navigational deskview
    // when called by gtk timeout at deskview_monitor.i. (recursive thread bumps
    // into a GDK_THREADS mutex). Anyways, there is no need to do by thread 
    // since there is no status area in deskview.


    if(view_p->flags.type == DESKVIEW_TYPE) {
	view_p->xfdir_p = load_xfdir_structure(view_p);
        rodent_init_grid (view_p, view_p->xfdir_p->pathc);
        load_xfdir_icons(view_p);
    } else {
	gchar *reading=g_strdup_printf(_("reading directory %s"), target_en->path);
	rfm_status (widgets_p, "xffm/stock_dialog-info", reading, NULL);
	gchar *text=NULL;
#ifdef ENABLE_READ_TIMEOUT
	gchar *text=g_strdup_printf("<b>%s</b>\n\n%s\n\n(<i>%s</i>)", 
		    //_("Waiting for Disk"),
		    _("Waiting for file list"),
		    reading,
		    _("The device is busy, waiting..."));
#endif
	TRACE("load_xfdir_structure\n");
	void *thread_result =
	    thread_tryjoin (load_xfdir_structure, view_p, text);
	g_free(reading);
	g_free(text);
	if (thread_result){
	    view_p->xfdir_p = thread_result;
	    TRACE ("hard_reload: st->st_uid  %d\n", 
			view_p->xfdir_p->st.st_uid );
	}
       	else {
	    // cancel and go back, if no back available, goto rodent-root
	    //rfm_destroy_entry(target_en);
	    GList *last = g_list_last (view_p->go_list);
	    if(!view_p->go_list || !last){
		target_en=NULL;
	    } else {
		target_en=rfm_copy_entry((record_entry_t *)last->data);
	    }
	    // harakiri to monitor thread:
	    
	    // I'm not sure this is not overkill:
	    if(target_en) view_p->module_name = target_en->module;
	    else view_p->module_name = NULL;
	    TRACE("going back...\n");

	    rfm_population_write_unlock (view_p);

	    goto retry_reload;
	}


        rodent_init_grid (view_p, view_p->xfdir_p->pathc);
        rodent_restore_view_size (view_p);
	// Thread 2: create the population (entries and icons)
        rfm_status (widgets_p, "xffm/stock_dialog-info", _("Loading folder..."), NULL);
        view_p->eyecandy.progress_count = 0;
        view_p->eyecandy.progress_total = view_p->xfdir_p->pathc;
        if(view_p->xfdir_p->pathc >= 512) {     /* eyecandy for long loads */
            set_progress_generic (&(view_p->widgets), 0, view_p->eyecandy.progress_total, 1);
        }
	// Main thread will be doing status bar stuff while
	// child thread does the work.
	//

	TRACE("load_xfdir_icons\n");
	thread_tryjoin (load_xfdir_icons, view_p, NULL);

        rfm_update_status_line (view_p);
        set_progress_generic (&(view_p->widgets), -1, -1, -1);
        rodent_set_scroll_position (view_p);
    } 
    g_free (widgets_p->workdir);
    widgets_p->workdir = g_strdup (target_en->path);

    if(view_p->en && view_p->en->st)
        memcpy (view_p->en->st, &view_p->xfdir_p->st, sizeof (struct stat));
    gdk_flush ();
 
    // fire off pending preview threads, (thread managed):
    // fire off pending preview threads, (thread managed):
    // but only if icons are larger than tiny
    // (tiny previews are for dummies)
    TRACE ("reload_iconview: rfm_fireup_previews\n");
    //if (view_p->view_layout.icon_size >= SMALL_ICON_SIZE)
    if (10){
	// since we are in barebones mode, this thread will also 
	// do the stat information filling in (fullbones)
	rfm_fireup_previews (view_p);
    }
    // resets light to green
    view_p->flags.status = STATUS_IDLE;

    // resaturate any icon which was previously saturated...
    const population_t *population_p = 
	rodent_find_in_population (view_p, view_p->mouse_event.current_mouseX, view_p->mouse_event.current_mouseY);
    if (population_p) {
	    rodent_saturate_item (view_p, population_p);	
    }

    rfm_cursor_reset (view_p->widgets.window);
    rfm_population_write_unlock (view_p);
    TRACE("unprotecting...\n");
    view_p->flags.protected=FALSE;
    // This following rodent_expose_all instruction is not threaded.
    rodent_expose_all (view_p);
	
    rfm_barrier_decrement();
    return TRUE;
}

#ifdef FULL_THREADED_LOAD
//
//  This is the proper way to do things (yet incomplete)
//             without borky calls to gtk_main_iteration().
//
//  XXX This will be a Rodent Delta milestone


int
rodent_full_reload_view (gpointer data, record_entry_t * target_en) {
    if (g_thread_self() != rfm_global_p->self){
	g_warning("Thread 0x%x: rodent_full_reload_view() may only be called from main thread (0x%x)", 
		GPOINTER_TO_INT(g_thread_self()), GPOINTER_TO_INT(rfm_global_p->self));
	return FALSE;
    }

    view_t *view_p = (view_t *) data;
    // XXX reload_in_progress should be a view_p variable, and all
    //     user events should be queued when this is on.
    static gboolean reload_in_progress=FALSE;
    if (reload_in_progress) {
	// XXX queue reload...
	g_warning("reload should be queued here...");
       return FALSE;	
    }
    full_reload_t *full_reload_p = (full_reload_t *)malloc(sizeof(full_reload_t));
    full_reload_p->view_p = view_p;
    full_reload_p->target_en = target_en;

    g_thread_create(rodent_full_reload_view_thread, full_reload_p, FALSE, NULL);

}
#endif

void
rodent_set_view_title (view_t * view_p) {
    if(view_p->flags.type == DESKVIEW_TYPE)
        return;
    TRACE (">> set_iconview_title\n");
    record_entry_t *en = view_p->en;
    const gchar *path;

    if(en && en->path && strlen (en->path)) {
        path = en->path;
    } else {
        path = g_get_host_name ();
    }

    if(view_p->module_name) {
        if(view_p->module_name && rfm_natural (PLUGIN_DIR, view_p->module_name, view_p->en, "window_title")) {
            path = rfm_natural (PLUGIN_DIR, view_p->module_name, view_p->en, "window_title");
        } else {
	    TRACE("module has no window_title function\n");
	}
    }
    set_icon_name (view_p->widgets.window, path);
    gchar *icon_id = NULL;

    if (rodent_path_has_bookmark(path)) {
	icon_id = g_strconcat(rfm_resolve_icon_id (en), "/composite/places_user-bookmarks", NULL);
    } else {
	icon_id = g_strdup(rfm_resolve_icon_id (en));
    }
    TRACE("icon resolved as %s\n", icon_id);

    set_application_icon (&(view_p->widgets), icon_id);

    gchar *basename = g_path_get_basename (path);
    gchar *dirname = g_path_get_dirname (path);
    gchar *parentname = g_path_get_basename (dirname);
    gchar *tabname;
    if(strcmp (parentname, ".") == 0 || strcmp (basename, "/") == 0) {
        tabname = g_strdup (basename);
    } else {
        tabname = g_build_filename (parentname, basename, NULL);
    }

    gchar *utf_tabname=rfm_utf_string(tabname);
    g_free(tabname);
    gchar *chopped_path=g_strdup(path);
    rfm_chop_excess (chopped_path);
    gchar *utf_path=rfm_utf_string(chopped_path);
    g_free(chopped_path);

   

    gtk_label_set_text (GTK_LABEL (view_p->widgets.page_label), utf_tabname);
    gtk_label_set_text (GTK_LABEL (view_p->widgets.page_menu_label), utf_path);
 
    g_free (basename);
    g_free (dirname);
    g_free (parentname);
    g_free (utf_tabname);
    GdkPixbuf *pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
    g_free(icon_id);
    GtkWidget *image = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image);
    GList *children = gtk_container_get_children (GTK_CONTAINER (view_p->widgets.page_icon));
    if(children) {
        gtk_container_remove (GTK_CONTAINER (view_p->widgets.page_icon), GTK_WIDGET (children->data));
    }
    gtk_container_add (GTK_CONTAINER (view_p->widgets.page_icon), image);
    g_list_free (children);

    return;
}


